import asyncio

import pytest

from asphalt.core.component import ContainerComponent, Component, component_types
from asphalt.core.context import Context


class DummyComponent(Component):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.started = False

    async def start(self, ctx):
        await asyncio.sleep(0.1)
        self.started = True


@pytest.fixture(autouse=True)
def monkeypatch_plugins(monkeypatch):
    monkeypatch.setattr(component_types, '_entrypoints', {'dummy': DummyComponent})


class TestContainerComponent:
    @pytest.fixture
    def container(self):
        return ContainerComponent({'dummy': {'a': 1, 'c': 3}})

    def test_add_component(self, container):
        """
        Test that add_component works with an without an entry point and that external
        configuration overriddes directly supplied configuration values.

        """
        container.add_component('dummy', DummyComponent, a=5, b=2)

        assert len(container.child_components) == 1
        component = container.child_components['dummy']
        assert isinstance(component, DummyComponent)
        assert component.kwargs == {'a': 1, 'b': 2, 'c': 3}

    @pytest.mark.parametrize('alias, cls, exc_cls, message', [
        ('', None, TypeError, 'component_alias must be a nonempty string'),
        ('foo', None, LookupError, 'no such entry point in asphalt.components: foo'),
        ('foo', int, TypeError,
         'int is not a subclass of asphalt.core.component.Component')
    ], ids=['empty_alias', 'bogus_entry_point', 'wrong_subclass'])
    def test_add_component_errors(self, container, alias, cls, exc_cls, message):
        exc = pytest.raises(exc_cls, container.add_component, alias, cls)
        assert str(exc.value) == message

    def test_add_duplicate_component(self, container):
        container.add_component('dummy')
        exc = pytest.raises(ValueError, container.add_component, 'dummy')
        assert str(exc.value) == 'there is already a child component named "dummy"'

    @pytest.mark.asyncio
    async def test_start(self, container):
        await container.start(Context())
        assert container.child_components['dummy'].started



from typing import Optional

import click
import yaml

from asphalt.core.component import component_types
from asphalt.core.runner import run_application, policies


@click.group()
def main():
    pass  # pragma: no cover


@main.command(help='Read a configuration file and start the application.')
@click.argument('configfile', type=click.File())
@click.option('--unsafe', is_flag=True, default=False,
              help='use unsafe mode when loading YAML (enables markup extensions)')
@click.option('-l', '--loop', type=click.Choice(policies.names),
              help='alternate event loop policy')
def run(configfile, unsafe: bool, loop: Optional[str]):
    # Read the configuration from the supplied YAML file
    config_data = yaml.load(configfile) if unsafe else yaml.safe_load(configfile)
    assert isinstance(config_data, dict), 'the document root element must be a dictionary'

    # Override the event loop policy if specified
    if loop:
        config_data['event_loop_policy'] = loop

    # Instantiate the root component
    try:
        component_config = config_data.pop('component')
    except KeyError:
        raise LookupError('missing configuration key: component') from None
    else:
        component = component_types.create_object(**component_config)

    # Start the application
    run_application(component, **config_data)


if __name__ == '__main__':  # pragma: no cover
    main()



import asyncio
from asyncio.futures import Future
from itertools import count

import pytest

from asphalt.core.context import ResourceConflict, ResourceNotFound, Resource, Context


class TestResource:
    @pytest.fixture
    def resource(self):
        return Resource(6, ('int', 'object'), 'foo', 'bar.foo')

    @pytest.fixture
    def lazy_resource(self):
        return Resource(None, ('int',), 'foo', 'bar.foo', lambda ctx: (ctx, 6))

    def test_get(self, resource):
        ctx = Context()
        assert resource.get_value(ctx) == 6

    def test_get_lazy(self):
        resource = Resource(None, ('int',), 'foo', None, lambda ctx: (ctx, 6))
        ctx = Context()
        assert resource.get_value(ctx) == (ctx, 6)

    def test_repr(self, resource: Resource):
        assert repr(resource) == ("Resource(types=('int', 'object'), alias='foo', "
                                  "value=6, context_attr='bar.foo', creator=None)")

    def test_str(self, resource: Resource):
        assert str(resource) == ("types=('int', 'object'), alias='foo', "
                                 "value=6, context_attr='bar.foo', creator=None")


class TestContext:
    @pytest.fixture
    def context(self):
        return Context(default_timeout=2)

    def test_parent(self):
        """Test that the parent property points to the parent context instance, if any."""
        parent = Context()
        child = Context(parent)
        assert parent.parent is None
        assert child.parent is parent

    @pytest.mark.asyncio
    async def test_contextmanager(self, context):
        """
        Test that "async with context:" dispatches both started and finished events and sets the
        exception variable when an exception is raised during the context lifetime.

        """
        events = []
        context.finished.connect(events.append)

        async with context:
            pass

        assert len(events) == 1

    @pytest.mark.asyncio
    async def test_contextmanager_exception(self, context):
        """
        Test that "async with context:" dispatches both started and finished events and sets the
        exception variable when an exception is raised during the context lifetime.

        """
        events = []
        context.finished.connect(events.append)

        with pytest.raises(RuntimeError) as exc:
            async with context:
                raise RuntimeError('test')

        assert len(events) == 1
        assert exc.value is events[0].exception

    @pytest.mark.asyncio
    async def test_publish_resource(self, context):
        """Test that a resource is properly published in the context and listeners are notified."""
        future = Future()
        context.resource_published.connect(future.set_result)
        context.publish_resource(6, 'foo', 'foo.bar', types=(int, float))

        value = await context.request_resource(int, 'foo')
        assert value == 6

        event = await future
        assert event.resource.types == ('int', 'float')
        assert event.resource.alias == 'foo'

    @pytest.mark.asyncio
    async def test_publish_resource_name_conflict(self, context):
        """Test that publishing a resource won't replace any existing resources."""
        context.publish_resource(5, 'foo')
        with pytest.raises(ResourceConflict) as exc:
            context.publish_resource(4, 'foo')

        assert str(exc.value) == (
            'this context has an existing resource of type int using the alias "foo"')

    @pytest.mark.asyncio
    async def test_publish_resource_context_attr(self, context):
        """
        Test that when resources are published, they are also set as properties of the context.

        Likewise, when they are removed, they are deleted from the context.
        """
        resource = context.publish_resource(1, context_attr='foo')
        assert context.foo == 1
        context.remove_resource(resource)
        assert 'foo' not in context.__dict__

    @pytest.mark.asyncio
    async def test_publish_resource_conflicting_attribute(self, context):
        """
        Test that the context won't allow publishing a resource with an attribute name that
        conflicts with an existing attribute.

        """
        context.a = 2
        with pytest.raises(ResourceConflict) as exc:
            context.publish_resource(2, context_attr='a')

        assert str(exc.value) == 'this context already has an attribute "a"'

        with pytest.raises(ResourceNotFound):
            await context.request_resource(int, timeout=0)

    @pytest.mark.parametrize('alias', ['a.b', 'a:b', 'a b'], ids=['dot', 'colon', 'space'])
    @pytest.mark.asyncio
    async def test_publish_resource_bad_alias(self, context, alias):
        with pytest.raises(AssertionError) as exc:
            context.publish_resource(1, alias)

        assert str(exc.value) == 'alias can only contain alphanumeric characters and underscores'

    @pytest.mark.asyncio
    async def test_publish_lazy_resource(self, context):
        """Test that lazy resources are only created once per context instance."""
        def creator(ctx):
            assert ctx is context
            return next(counter)

        counter = count(1)
        context.publish_lazy_resource(creator, int, context_attr='foo')
        assert context.foo == 1
        assert context.foo == 1
        assert context.__dict__['foo'] == 1

    @pytest.mark.asyncio
    async def test_publish_lazy_resource_coroutine(self, context):
        """
        Test that a coroutine function can be a resource creator and that it returns a Future.

        """
        async def async_creator(context):
            return 'hello'

        context.publish_lazy_resource(async_creator, str, 'foo', 'foo')
        hello = context.foo
        assert isinstance(hello, Future)
        assert await context.request_resource(str, 'foo') == 'hello'

    @pytest.mark.asyncio
    async def test_publish_lazy_resource_coroutine_delayed(self, event_loop, context):
        """
        Test that a coroutine resource creator works even when the resource isn't immediately
        available.

        """
        async def async_creator(context):
            return 'hello'

        future = event_loop.create_task(context.request_resource(str, 'foo'))
        await asyncio.sleep(0.2)
        context.publish_lazy_resource(async_creator, str, alias='foo')
        assert await future == 'hello'

    @pytest.mark.asyncio
    async def test_publish_lazy_resource_conflicting_resource(self, context):
        context.publish_lazy_resource(lambda ctx: 2, int, context_attr='a')
        with pytest.raises(ResourceConflict) as exc:
            context.publish_resource(2, 'foo', context_attr='a')

        assert str(exc.value) == (
            'this context has an existing lazy resource using the attribute "a"')

    @pytest.mark.asyncio
    async def test_publish_lazy_resource_duplicate(self, context):
        context.publish_lazy_resource(lambda ctx: None, str, context_attr='foo')
        with pytest.raises(ResourceConflict) as exc:
            await context.publish_lazy_resource(lambda ctx: None, str, context_attr='foo')

        assert (str(exc.value) ==
                'this context has an existing resource of type str using the alias "default"')

    @pytest.mark.asyncio
    async def test_remove_resource(self, context):
        """Test that resources can be removed and that the listeners are notified."""
        future = Future()
        context.resource_removed.connect(future.set_result)
        resource = context.publish_resource(4)
        context.remove_resource(resource)

        event = await future
        assert event.resource.types == ('int',)

        with pytest.raises(ResourceNotFound):
            await context.request_resource(int, timeout=0)

    @pytest.mark.asyncio
    async def test_remove_nonexistent(self, context):
        resource = Resource(5, ('int',), 'default', None)
        with pytest.raises(LookupError) as exc:
            context.remove_resource(resource)

        assert str(exc.value) == ("Resource(types=('int',), alias='default', value=5, "
                                  "context_attr=None, creator=None) not found in this context")

    @pytest.mark.asyncio
    async def test_request_timeout(self, context):
        with pytest.raises(ResourceNotFound) as exc:
            await context.request_resource(int, timeout=0.2)

        assert str(exc.value) == "no matching resource was found for type='int' alias='default'"

    @pytest.mark.asyncio
    @pytest.mark.parametrize('bad_arg, errormsg', [
        ('type', 'type must be a type or a nonempty string'),
        ('alias', 'alias must be a nonempty string')
    ], ids=['bad_type', 'bad_alias'])
    async def test_bad_request(self, context, bad_arg, errormsg):
        type_ = '' if bad_arg == 'type' else 'foo'
        alias = '' if bad_arg == 'alias' else 'foo'
        with pytest.raises(ValueError) as exc:
            await context.request_resource(type_, alias)
        assert str(exc.value) == errormsg

    def test_attribute_error(self, context):
        exc = pytest.raises(AttributeError, getattr, context, 'foo')
        assert str(exc.value) == 'no such context variable: foo'

    def test_get_parent_attribute(self, context):
        """
        Test that accessing a nonexistent attribute on a context retrieves the value from parent.

        """
        child_context = Context(context)
        context.a = 2
        assert child_context.a == 2

    @pytest.mark.asyncio
    async def test_request_resource_parent_add(self, context, event_loop):
        """
        Test that publishing a resource to the parent context will satisfy a resource request in a
        child context.

        """
        child_context = Context(context)
        task = event_loop.create_task(child_context.request_resource(int))
        context.publish_resource(6)
        resource = await task
        assert resource == 6

    @pytest.mark.asyncio
    async def test_request_lazy_resource_context_attr(self, context):
        """Test that requesting a lazy resource also sets the context variable."""
        context.publish_lazy_resource(lambda ctx: 6, int, context_attr='foo')
        await context.request_resource(int)
        assert context.__dict__['foo'] == 6

    @pytest.mark.asyncio
    async def test_remove_lazy_resource(self, context):
        """
        Test that the lazy resource is no longer created when it has been removed and its
        context variable is accessed.

        """
        resource = context.publish_lazy_resource(lambda ctx: 6, int, context_attr='foo')
        context.remove_resource(resource)
        exc = pytest.raises(AttributeError, getattr, context, 'foo')
        assert str(exc.value) == 'no such context variable: foo'

    @pytest.mark.asyncio
    async def test_get_resources(self, context):
        resource1 = context.publish_resource(6, 'int1')
        resource2 = context.publish_resource(8, 'int2')
        resource3 = context.publish_resource('foo', types=(str, 'collections.abc.Iterable'))
        resource4 = context.publish_resource(
            (5, 4), 'sometuple', types=(tuple, 'collections.abc.Iterable'))

        assert context.get_resources() == [resource1, resource2, resource3, resource4]
        assert context.get_resources(int) == [resource1, resource2]
        assert context.get_resources(str) == [resource3]
        assert context.get_resources('collections.abc.Iterable') == [resource3, resource4]

    @pytest.mark.asyncio
    async def test_get_resources_include_parents(self, context):
        subcontext = Context(context)
        resource1 = context.publish_resource(6, 'int1')
        resource2 = subcontext.publish_resource(8, 'int2')
        resource3 = context.publish_resource('foo', 'str')

        assert subcontext.get_resources() == [resource1, resource2, resource3]
        assert subcontext.get_resources(include_parents=False) == [resource2]



import logging
from asyncio import Future, get_event_loop, Queue
from datetime import datetime, timezone
from inspect import isawaitable, iscoroutine, getmembers
from time import time, monotonic
from traceback import format_exception
from types import CoroutineType
from typing import Callable, Any, Sequence, Tuple, Optional
from weakref import WeakKeyDictionary

from asyncio_extras.asyncyield import yield_async
from asyncio_extras.generator import async_generator
from typeguard import check_argument_types

from asphalt.core.util import qualified_name

__all__ = ('Event', 'EventDispatchError', 'Signal', 'wait_event', 'stream_events')

logger = logging.getLogger(__name__)


class Event:
    """
    The base class for all events.

    :param source: the object where this event originated from
    :param topic: the event topic

    :ivar source: the object where this event was dispatched from
    :ivar str topic: the topic
    :ivar float time: event creation time as seconds from the UNIX epoch
    :ivar float monotime: event creation time, as returned by :func:`time.monotonic`
    """

    __slots__ = 'source', 'topic', 'time', 'monotime'

    def __init__(self, source, topic: str):
        self.source = source
        self.topic = topic
        self.time = time()
        self.monotime = monotonic()

    @property
    def utc_timestamp(self) -> datetime:
        """
        Return a timezone aware :class:`~datetime.datetime` corresponding to the ``time`` variable,
        using the UTC timezone.

        """
        return datetime.fromtimestamp(self.time, timezone.utc)

    def __repr__(self):
        return '{0.__class__.__name__}(source={0.source!r}, topic={0.topic!r})'.format(self)


class EventDispatchError(Exception):
    """
    Raised when one or more event listener callbacks raise an exception.

    The tracebacks of all the exceptions are displayed in the exception message.

    :ivar Event event: the event
    :ivar exceptions: a sequence containing tuples of (callback, exception) for each exception that
        was raised by a listener callback
    :vartype exceptions: Sequence[Tuple[Callable, Exception]]
    """

    def __init__(self, event: Event, exceptions: Sequence[Tuple[Callable, Exception]]):
        message = '-------------------------------\n'.join(
            ''.join(format_exception(type(exc), exc, exc.__traceback__)) for _, exc in exceptions)
        super().__init__('error dispatching event:\n' + message)
        self.event = event
        self.exceptions = exceptions


class Signal:
    """
    Declaration of a signal that can be used to dispatch events.

    This is a descriptor that returns itself on class level attribute access and a bound version of
    itself on instance level access. Connecting listeners and dispatching events only works with
    these bound instances.

    Each signal must be assigned to a class attribute, but only once. Assigning the same Signal
    instance to more than one attribute will raise a :exc:`LookupError` on attribute access.

    :param event_class: an event class
    """
    __slots__ = 'event_class', 'source', 'topic', 'listeners', 'bound_signals'

    def __init__(self, event_class: type, *, source=None, topic: str = None):
        assert check_argument_types()
        assert issubclass(event_class, Event), 'event_class must be a subclass of Event'
        self.event_class = event_class
        self.topic = topic
        if source is not None:
            self.source = source
            self.listeners = []
        else:
            self.bound_signals = WeakKeyDictionary()

    def __get__(self, instance, owner) -> 'Signal':
        if instance is None:
            return self

        # Find the attribute this Signal was assigned to
        if self.topic is None:
            attrnames = [attr for attr, value in getmembers(owner, lambda value: value is self)]
            if len(attrnames) > 1:
                raise LookupError('this Signal was assigned to multiple attributes: ' +
                                  ', '.join(attrnames))
            else:
                self.topic = attrnames[0]

        try:
            return self.bound_signals[instance]
        except KeyError:
            bound_signal = Signal(self.event_class, source=instance, topic=self.topic)
            self.bound_signals[instance] = bound_signal
            return bound_signal

    def connect(self, callback: Callable[[Event], Any]) -> Callable[[Event], Any]:
        """
        Connect a callback to this signal.

        Each callable can only be connected once. Duplicate registrations are ignored.

        If you need to pass extra arguments to the callback, you can use :func:`functools.partial`
        to wrap the callable.

        :param callback: a callable that will receive an event object as its only argument.
        :return: the value of ``callback`` argument

        """
        assert check_argument_types()
        if callback not in self.listeners:
            self.listeners.append(callback)

        return callback

    def disconnect(self, callback: Callable) -> None:
        """
        Disconnects the given callback.

        The callback will no longer receive events from this signal.

        No action is taken if the callback is not on the list of listener callbacks.

        :param callback: the callable to remove

        """
        assert check_argument_types()
        try:
            self.listeners.remove(callback)
        except ValueError:
            pass

    def dispatch_event(self, event: Event, *, return_future: bool = False) -> Optional[Future]:
        """
        Dispatch the given event to all listeners.

        Creates a new task in which all listener callbacks are called with the given event as
        the only argument. Coroutine callbacks are converted to their own respective tasks and
        waited for concurrently.

        :param event: the event object to dispatch
        :param return_future:
            If ``True``, return a :class:`~asyncio.Future` that completes when all the listener
            callbacks have been processed. If any one of them raised an exception, the future will
            have an :exc:`~asphalt.core.event.EventDispatchError` exception set in it which
            contains all of the exceptions raised in the callbacks.

            If set to ``False``, then ``None`` will be returned, and any exceptions raised in
            listener callbacks will be logged instead.
        :returns: a future or ``None``, depending on the ``return_future`` argument

        """
        async def do_dispatch():
            futures, exceptions = [], []
            for callback in listeners:
                try:
                    retval = callback(event)
                except Exception as e:
                    exceptions.append((callback, e))
                    if not return_future:
                        logger.exception('uncaught exception in event listener')
                else:
                    if isawaitable(retval):
                        if iscoroutine(retval):
                            retval = loop.create_task(retval)

                        futures.append((callback, retval))

            # For any callbacks that returned awaitables, wait for their completion and collect any
            # exceptions they raise
            for callback, awaitable in futures:
                try:
                    await awaitable
                except Exception as e:
                    exceptions.append((callback, e))
                    if not return_future:
                        logger.exception('uncaught exception in event listener')

            if exceptions and return_future:
                raise EventDispatchError(event, exceptions)

        assert isinstance(event, self.event_class), \
            'event must be of type {}'.format(qualified_name(self.event_class))
        if self.listeners:
            loop = get_event_loop()
            listeners = self.listeners.copy()
            future = loop.create_task(do_dispatch())
            return future if return_future else None
        elif return_future:
            # The event has no listeners, so skip the task creation and return an empty Future
            future = Future()
            future.set_result(None)
            return future
        else:
            return None

    def dispatch(self, *args, return_future: bool = False, **kwargs) -> Optional[Future]:
        """
        Create and dispatch an event.

        This method constructs an event object and then passes it to :meth:`dispatch_event` for
        the actual dispatching.

        :param args: positional arguments to the constructor of the associated event class.
        :param return_future:
            If ``True``, return a :class:`~asyncio.Future` that completes when all the listener
            callbacks have been processed. If any one of them raised an exception, the future will
            have an :exc:`~asphalt.core.event.EventDispatchError` exception set in it which
            contains all of the exceptions raised in the callbacks.

            If set to ``False``, then ``None`` will be returned, and any exceptions raised in
            listener callbacks will be logged instead.
        :returns: a future or ``None``, depending on the ``return_future`` argument

        """
        assert check_argument_types()
        event = self.event_class(self.source, self.topic, *args, **kwargs)
        return self.dispatch_event(event, return_future=return_future)

    def wait_event(self) -> CoroutineType:
        """Shortcut for calling :func:`wait_event` with this signal as the argument."""
        return wait_event(self)

    def stream_events(self, max_queue_size: int = 0):
        """Shortcut for calling :func:`stream_events` with this signal as the argument."""
        return stream_events(self, max_queue_size=max_queue_size)


async def wait_event(*signals: Signal) -> Event:
    """Return the first event dispatched from any of the given signals."""
    future = Future()
    for signal in signals:
        signal.connect(future.set_result)

    try:
        return await future
    finally:
        for signal in signals:
            signal.disconnect(future.set_result)


@async_generator
async def stream_events(*signals: Signal, max_queue_size: int = 0):
    """
    Generate event objects to the consumer as they're dispatched.

    This function is meant for use with ``async for``.

    :param signals: one or more signals to get events from
    :param max_queue_size: maximum size of the queue, after which it will start to drop events

    """
    queue = Queue(max_queue_size)
    for signal in signals:
        signal.connect(queue.put_nowait)

    try:
        while True:
            event = await queue.get()
            await yield_async(event)
    finally:
        for signal in signals:
            signal.disconnect(queue.put_nowait)



import asyncio
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from typing import Dict, Any, Union

from typeguard import check_argument_types

from asphalt.core.context import Context
from asphalt.core.util import PluginContainer, merge_config

__all__ = ('Component', 'ContainerComponent')


class Component(metaclass=ABCMeta):
    """This is the base class for all Asphalt components."""

    __slots__ = ()

    @abstractmethod
    async def start(self, ctx: Context):
        """
        Perform any necessary tasks to start the services provided by this component.

        Components typically use the context to:
          * publish resources (:meth:`~asphalt.core.context.Context.publish_resource` and
            :meth:`~asphalt.core.context.Context.publish_lazy_resource`)
          * request resources (:meth:`~asphalt.core.context.Context.request_resource`)

        It is advisable for Components to first publish all the resources they can before
        requesting any. This will speed up the dependency resolution and prevent deadlocks.

        :param ctx: the containing context for this component
        """


class ContainerComponent(Component):
    """
    A component that can contain other components.

    :ivar child_components: dictionary of component alias ⭢ :class:`Component` instance (of child
        components added with :meth:`add_component`)
    :vartype child_components: Dict[str, Component]
    :ivar component_configs: dictionary of component alias ⭢ externally provided component
        configuration
    :vartype component_configs: Dict[str, Dict[str, Any]]
    """

    __slots__ = 'child_components', 'component_configs'

    def __init__(self, components: Dict[str, Dict[str, Any]] = None):
        assert check_argument_types()
        self.child_components = OrderedDict()
        self.component_configs = components or {}

    def add_component(self, alias: str, type: Union[str, type] = None, **config):
        """
        Add a child component.

        This will instantiate a component class, as specified by the ``type`` argument.

        If the second argument is omitted, the value of ``alias`` is used as its value.

        The locally given configuration can be overridden by component configuration parameters
        supplied to the constructor (via the ``components`` argument).

        When configuration values are provided both as keyword arguments to this method and
        component configuration through the ``components`` constructor argument, the configurations
        are merged together using :func:`~asphalt.core.util.merge_config` in a way that the
        configuration values from the ``components`` argument override the keyword arguments to
        this method.

        :param alias: a name for the component instance, unique within this container
        :param type: entry point name or :class:`Component` subclass or a ``module:varname``
            reference to one
        :param config: keyword arguments passed to the component's constructor

        """
        assert check_argument_types()
        if not isinstance(alias, str) or not alias:
            raise TypeError('component_alias must be a nonempty string')
        if alias in self.child_components:
            raise ValueError('there is already a child component named "{}"'.format(alias))

        # Allow the external configuration to override the constructor arguments
        override_config = self.component_configs.get(alias) or {}
        config = merge_config(config, override_config)

        component = component_types.create_object(type or alias, **config)
        self.child_components[alias] = component

    async def start(self, ctx: Context):
        """
        Create child components that have been configured but not yet created and then calls their
        :meth:`~Component.start` methods in separate tasks and waits until they have completed.

        """
        for alias in self.component_configs:
            if alias not in self.child_components:
                self.add_component(alias)

        loop = asyncio.get_event_loop()
        tasks = [loop.create_task(component.start(ctx)) for component in
                 self.child_components.values()]
        if tasks:
            await asyncio.gather(*tasks)


component_types = PluginContainer('asphalt.components', Component)



import asyncio
import sys
from concurrent.futures import ThreadPoolExecutor
from logging import basicConfig, getLogger, INFO
from logging.config import dictConfig
from typing import Union, Dict, Any

from typeguard import check_argument_types

from asphalt.core.component import Component
from asphalt.core.context import Context
from asphalt.core.util import PluginContainer, qualified_name

__all__ = ('run_application',)

policies = PluginContainer('asphalt.core.event_loop_policies')


def uvloop_policy():
    import uvloop
    return uvloop.EventLoopPolicy()


def gevent_policy():
    import aiogevent
    return aiogevent.EventLoopPolicy()


def run_application(component: Component, *, event_loop_policy: str = None,
                    max_threads: int = None, logging: Union[Dict[str, Any], int, None] = INFO):
    """
    Configure logging and start the given root component in the default asyncio event loop.

    Assuming the root component was started successfully, the event loop will continue running
    until the process is terminated.

    Initializes the logging system first based on the value of ``logging``:
      * If the value is a dictionary, it is passed to :func:`logging.config.dictConfig` as
        argument.
      * If the value is an integer, it is passed to :func:`logging.basicConfig` as the logging
        level.
      * If the value is ``None``, logging setup is skipped entirely.

    By default, the logging system is initialized using :func:`~logging.basicConfig` using the
    ``INFO`` logging level.

    The default executor in the event loop is replaced with a new
    :class:`~concurrent.futures.ThreadPoolExecutor` where the maximum number of threads is set to
    the value of ``max_threads`` or, if omitted, the return value of :func:`os.cpu_count()`.

    :param component: the root component
    :param event_loop_policy: entry point name (from the ``asphalt.core.event_loop_policies``
        namespace) of an alternate event loop policy (or a module:varname reference to one)
    :param max_threads: the maximum number of worker threads in the default thread pool executor
        (the default value depends on the event loop implementation)
    :param logging: a logging configuration dictionary, :ref:`logging level <python:levels>` or
        ``None``

    """
    assert check_argument_types()

    # Configure the logging system
    if isinstance(logging, dict):
        dictConfig(logging)
    elif isinstance(logging, int):
        basicConfig(level=logging)

    # Switch to an alternate event loop policy if one was provided
    logger = getLogger(__name__)
    if event_loop_policy:
        create_policy = policies.resolve(event_loop_policy)
        policy = create_policy()
        asyncio.set_event_loop_policy(policy)
        logger.info('Switched event loop policy to %s', qualified_name(policy))

    # Assign a new default executor with the given max worker thread limit if one was provided
    event_loop = asyncio.get_event_loop()
    if max_threads is not None:
        event_loop.set_default_executor(ThreadPoolExecutor(max_threads))
        logger.info('Installed a new thread pool executor with max_workers=%d', max_threads)

    logger.info('Starting application')
    context = Context()
    exception = None
    try:
        try:
            event_loop.run_until_complete(component.start(context))
        except Exception as e:
            exception = e
            logger.exception('Error during application startup')
        else:
            # Enable garbage collection of the component tree
            del component

            # Finally, run the event loop until the process is terminated or Ctrl+C is pressed
            event_loop.run_forever()
    except (KeyboardInterrupt, SystemExit):
        pass
    finally:
        # Cancel all running tasks
        for task in asyncio.Task.all_tasks(event_loop):
            task.cancel()

        # Run all the finish callbacks
        future = context.finished.dispatch(exception, return_future=True)
        event_loop.run_until_complete(future)

    event_loop.close()
    logger.info('Application stopped')

    if exception is not None:
        sys.exit(1)



import asyncio
from asyncio.futures import Future
from datetime import datetime, timezone

import pytest

from asphalt.core.event import Event, EventDispatchError, Signal, stream_events, wait_event


class DummyEvent(Event):
    def __init__(self, source, topic, *args, **kwargs):
        super().__init__(source, topic)
        self.args = args
        self.kwargs = kwargs


class DummySource:
    event_a = Signal(DummyEvent)
    event_b = Signal(DummyEvent)


@pytest.fixture
def source():
    return DummySource()


class TestEvent:
    def test_utc_timestamp(self, source):
        event = Event(source, 'sometopic')
        assert isinstance(event.utc_timestamp, datetime)
        assert event.utc_timestamp.tzinfo == timezone.utc

    def test_event_repr(self, source):
        event = Event(source, 'sometopic')
        assert repr(event) == "Event(source=%r, topic='sometopic')" % source


class TestSignal:
    def test_class_attribute_access(self):
        """
        Test that accessing the descriptor on the class level returns the same signal instance.

        """
        signal = Signal(DummyEvent)

        class EventSource:
            dummysignal = signal

        assert EventSource.dummysignal is signal

    def test_multi_attribute_assignment(self):
        """
        Test that assigning the signal object to multiple class attributes raises a ``LookupError``
        on instance attribute access.

        """
        signal = Signal(DummyEvent)

        class EventSource:
            dummysignal = signal
            dummysignal2 = signal

        with pytest.raises(LookupError) as exc:
            getattr(EventSource(), 'dummysignal')

        assert str(exc.value) == \
            'this Signal was assigned to multiple attributes: dummysignal, dummysignal2'

    @pytest.mark.asyncio
    async def test_disconnect(self, source):
        """Test that an event listener no longer receives events after it's been removed."""
        events = []
        source.event_a.connect(events.append)
        await source.event_a.dispatch(1, return_future=True)

        source.event_a.disconnect(events.append)
        await source.event_a.dispatch(2, return_future=True)

        assert len(events) == 1
        assert events[0].args == (1,)

    def test_remove_nonexistent_listener(self, source):
        """Test that attempting to remove a nonexistent event listener will not raise an error."""
        source.event_a.disconnect(lambda: None)

    @pytest.mark.asyncio
    async def test_dispatch_event_coroutine(self, source):
        """Test that a coroutine function can be an event listener."""
        async def callback(event: Event):
            events.append(event)

        events = []
        source.event_a.connect(callback)
        await source.event_a.dispatch('x', 'y', return_future=True, a=1, b=2)

        assert len(events) == 1
        assert events[0].args == ('x', 'y')
        assert events[0].kwargs == {'a': 1, 'b': 2}

    @pytest.mark.asyncio
    async def test_dispatch_event(self, source):
        """Test that dispatch_event() correctly dispatches the given event."""
        events = []
        source.event_a.connect(events.append)
        event = DummyEvent(source, 'event_a', 'x', 'y', a=1, b=2)
        await source.event_a.dispatch_event(event, return_future=True)

        assert events == [event]

    @pytest.mark.asyncio
    async def test_dispatch_listener_exception_logging(self, event_loop, source, caplog):
        """Test that listener exceptions are logged when return_future is False."""
        def listener(event):
            try:
                raise Exception('regular')
            finally:
                future1.set_result(None)

        async def coro_listener(event):
            try:
                raise Exception('coroutine')
            finally:
                future2.set_result(None)

        future1, future2 = Future(), Future()
        source.event_a.connect(listener)
        source.event_a.connect(coro_listener)
        source.event_a.dispatch()
        await asyncio.gather(future1, future2)

        assert len(caplog.records) == 2
        for record in caplog.records:
            assert 'uncaught exception in event listener' in record.message

    def test_dispatch_event_no_listeners(self, source):
        """Test that None is returned when no listeners are present and return_future is False."""
        assert source.event_a.dispatch() is None

    @pytest.mark.asyncio
    async def test_dispatch_event_listener_exceptions(self, source):
        """
        Test that multiple exceptions raised by listeners are combined into one EventDispatchError.

        """
        def listener(event):
            raise plain_exception

        async def coro_listener(event):
            raise async_exception

        plain_exception = ValueError('foo')
        async_exception = KeyError('bar')
        plain_listener = source.event_a.connect(listener)
        async_listener = source.event_a.connect(coro_listener)
        event = DummyEvent(source, 'event_a')
        with pytest.raises(EventDispatchError) as exc:
            await source.event_a.dispatch_event(event, return_future=True)

        assert exc.value.event is event
        assert exc.value.exceptions == [
            (plain_listener, plain_exception),
            (async_listener, async_exception)
        ]
        assert 'listener' in str(exc.value)
        assert 'coro_listener' in str(exc.value)
        assert '-------------------------------\n' in str(exc.value)

    @pytest.mark.asyncio
    async def test_dispatch_event_class_mismatch(self, source):
        """Test that passing an event of the wrong type raises an AssertionError."""
        with pytest.raises(AssertionError) as exc:
            await source.event_a.dispatch_event(Event(source, 'event_a'))
        assert str(exc.value) == 'event must be of type test_event.DummyEvent'

    @pytest.mark.asyncio
    async def test_wait_event(self, source, event_loop):
        task = event_loop.create_task(source.event_a.wait_event())
        event_loop.call_soon(source.event_a.dispatch)
        received_event = await task
        assert received_event.topic == 'event_a'

    @pytest.mark.asyncio
    async def test_stream_events(self, source, event_loop):
        async def generate_events():
            await asyncio.sleep(0.1)
            source.event_a.dispatch(1)
            await asyncio.sleep(0.1)
            source.event_a.dispatch(2)
            await asyncio.sleep(0.1)
            source.event_a.dispatch(3)

        event_loop.create_task(generate_events())
        last_number = 0
        async for event in source.event_a.stream_events():
            assert event.args[0] == last_number + 1
            last_number += 1
            if last_number == 3:
                break


@pytest.mark.asyncio
async def test_wait_event(event_loop):
    """Test that wait_event catches events coming from any of the given signals."""
    source1, source2 = DummySource(), DummySource()

    for signal in source1.event_a, source2.event_b:
        task = event_loop.create_task(wait_event(signal))
        event_loop.call_soon(signal.dispatch)
        received_event = await task
        assert received_event.topic == signal.topic


@pytest.mark.asyncio
async def test_stream_events(event_loop):
    async def generate_events():
        await asyncio.sleep(0.1)
        source1.event_a.dispatch(1)
        await asyncio.sleep(0.1)
        source2.event_b.dispatch(2)
        await asyncio.sleep(0.1)
        source1.event_a.dispatch(3)

    source1, source2 = DummySource(), DummySource()
    event_loop.create_task(generate_events())
    last_number = 0
    async for event in stream_events(source1.event_a, source2.event_b):
        assert event.args[0] == last_number + 1
        last_number += 1
        if last_number == 3:
            break



import re
from asyncio import Future, TimeoutError, wait_for, ensure_future
from inspect import isawaitable, iscoroutine
from itertools import chain
from typing import Optional, Callable, Any, Union, Iterable, Sequence, Dict  # noqa

from collections import defaultdict
from typeguard import check_argument_types

from asphalt.core.event import Signal, Event
from asphalt.core.util import qualified_name

__all__ = ('Resource', 'ResourceEvent', 'ResourceConflict', 'ResourceNotFound',
           'ContextFinishEvent', 'Context')


class Resource:
    """
    Contains the resource value or its creator callable, plus some metadata.

    :ivar str alias: alias of the resource
    :ivar Sequence[str] types: type names the resource was registered with
    :ivar str context_attr: the context attribute of the resource
    :ivar creator: callable to create the value (in case of a lazy resource)
    :vartype creator: Callable[[Context], Any]
    """

    __slots__ = 'value', 'types', 'alias', 'context_attr', 'creator'

    def __init__(self, value, types: Sequence[str], alias: str, context_attr: Optional[str],
                 creator: Callable[['Context'], Any] = None):
        assert check_argument_types()
        self.value = value
        self.types = types
        self.alias = alias
        self.context_attr = context_attr
        self.creator = creator

    def get_value(self, ctx: 'Context'):
        assert check_argument_types()
        if self.value is None and self.creator is not None:
            self.value = self.creator(ctx)
            if iscoroutine(self.value):
                self.value = ensure_future(self.value)
            if self.context_attr:
                setattr(ctx, self.context_attr, self.value)

        return self.value

    def __repr__(self):
        return '{0.__class__.__name__}({0})'.format(self)

    def __str__(self):
        return ('types={0.types!r}, alias={0.alias!r}, value={0.value!r}, '
                'context_attr={0.context_attr!r}, creator={1}'
                .format(self, qualified_name(self.creator) if self.creator else None))


class ResourceEvent(Event):
    """
    Dispatched when a resource has been published to or removed from a context.

    :ivar Context source: the relevant context
    :ivar Resource resource: the resource that was published or removed
    """

    __slots__ = 'resource'

    def __init__(self, source: 'Context', topic: str, resource: Resource):
        assert check_argument_types()
        super().__init__(source, topic)
        self.resource = resource


class ResourceConflict(Exception):
    """
    Raised when a new resource that is being published conflicts with an existing resource or
    context variable.
    """


class ResourceNotFound(LookupError):
    """Raised when a resource request cannot be fulfilled within the allotted time."""

    def __init__(self, type: str, alias: str):
        assert check_argument_types()
        super().__init__(type, alias)
        self.type = type
        self.alias = alias

    def __str__(self):
        return 'no matching resource was found for type={0.type!r} alias={0.alias!r}'.format(self)


class ContextFinishEvent(Event):
    """
    Dispatched when a context has served its purpose and is being torn down.

    :ivar Optional[BaseException] exception: the exception that caused the context to finish, if
        any
    """

    __slots__ = 'exception'

    def __init__(self, source: 'Context', topic: str, exception: Optional[BaseException]):
        assert check_argument_types()
        super().__init__(source, topic)
        self.exception = exception


class Context:
    """
    Contexts give request handlers and callbacks access to resources.

    Contexts are stacked in a way that accessing an attribute that is not present in the current
    context causes the attribute to be looked up in the parent instance and so on, until the
    attribute is found (or :class:`AttributeError` is raised).

    Requesting or publishing of resources **MUST NOT** be attempted during or after the dispatch
    of the ``finished`` event.

    :param parent: the parent context, if any
    :param default_timeout: default timeout for :meth:`request_resource` if omitted from the call
        arguments

    :var Signal finished: a signal (:class:`ContextFinishEvent`) dispatched when the context has
        served its purpose and is being discarded
    :var Signal resource_published: a signal (:class:`ResourceEvent`) dispatched when a resource
        has been published in this context
    :var Signal resource_removed: a signal (:class:`ResourceEvent`): dispatched when a resource has
        been removed from this context
    """

    finished = Signal(ContextFinishEvent)
    resource_published = Signal(ResourceEvent)
    resource_removed = Signal(ResourceEvent)

    def __init__(self, parent: 'Context' = None, *, default_timeout: int = 5):
        assert check_argument_types()
        self._parent = parent
        self._resources = defaultdict(dict)  # type: Dict[str, Dict[str, Resource]]
        self._lazy_resources = {}  # type: Dict[str, Resource]
        self.default_timeout = default_timeout

    def __getattr__(self, name):
        resource = self._lazy_resources.get(name)
        if resource is not None:
            value = resource.get_value(self)
            setattr(self, name, value)
            return value

        if self._parent is not None:
            return getattr(self._parent, name)

        raise AttributeError('no such context variable: {}'.format(name))

    @property
    def parent(self) -> Optional['Context']:
        """Return the parent of this context or ``None`` if there is no parent context."""
        return self._parent

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.finished.dispatch(exc_val, return_future=True)

    def _publish_resource(self, value, alias: str, context_attr: str,
                          types: Iterable[Union[str, type]],
                          creator: Optional[Callable[['Context'], Any]]):
        assert isinstance(alias, str) and alias, 'alias must be a nonempty string'
        assert re.match(r'^\w+$', alias),\
            'alias can only contain alphanumeric characters and underscores'
        assert context_attr is None or isinstance(context_attr, str),\
            'context_attr must be a nonempty string or None'

        if isinstance(types, (str, type)):
            types = (types,)

        # Check for alias conflicts with existing resources
        types = tuple(t if isinstance(t, str) else qualified_name(t) for t in types)
        for typename in types:
            if alias in self._resources[typename]:
                raise ResourceConflict(
                    'this context has an existing resource of type {} using the alias "{}"'
                    .format(typename, alias))

        # Check for context attribute conflicts
        if context_attr:
            # Check that there is no existing attribute by that name
            if context_attr in dir(self):
                raise ResourceConflict(
                    'this context already has an attribute "{}"'.format(context_attr))

            # Check that there is no existing lazy resource using the
            # same context attribute
            if context_attr in self._lazy_resources:
                raise ResourceConflict(
                    'this context has an existing lazy resource using the attribute "{}"'
                    .format(context_attr))

        # Register the resource
        resource = Resource(value, types, alias, context_attr, creator)
        for typename in types:
            self._resources[typename][resource.alias] = resource

        if creator is not None and context_attr is not None:
            self._lazy_resources[context_attr] = resource

        # Add the resource as an attribute of this context if context_attr is defined
        if creator is None and resource.context_attr:
            setattr(self, context_attr, value)

        self.resource_published.dispatch(resource)
        return resource

    def publish_resource(
            self, value, alias: str = 'default', context_attr: str = None, *,
            types: Union[Union[str, type], Iterable[Union[str, type]]] = ()) -> Resource:
        """
        Publish a resource and dispatch a ``resource_published``  event.

        :param value: the actual resource value
        :param alias: name of this resource (unique among all its registered types)
        :param context_attr: name of the context attribute this resource will be accessible as
        :param types: type(s) to register the resource as (omit to use the type of ``value``)
        :return: the resource handle
        :raises asphalt.core.context.ResourceConflict: if the resource conflicts with an existing
            one in any way

        """
        assert check_argument_types()
        assert value is not None, 'value must not be None'
        if not types:
            types = [type(value)]

        return self._publish_resource(value, alias, context_attr, types, None)

    def publish_lazy_resource(self, creator: Callable[['Context'], Any],
                              types: Union[Union[str, type], Iterable[Union[str, type]]],
                              alias: str = 'default', context_attr: str = None) -> Resource:
        """
        Publish a "lazy" or "contextual" resource and dispatch a ``resource_published`` event.

        Instead of a concrete resource value, you supply a creator callable which is called with a
        context object as its argument when the resource is being requested either via
        :meth:`request_resource` or by context attribute access.
        The return value of the creator callable will be cached so the creator will only be called
        once per context instance.

        If the creator callable is a coroutine function or returns an awaitable, it is resolved
        before storing the resource value and returning it to the requester. Note that this will
        **NOT** work when a context attribute has been specified for the resource.

        :param creator: a callable taking a context instance as argument
        :param types: type(s) to register the resource as
        :param alias: name of this resource (unique among all its registered types)
        :param context_attr: name of the context attribute this resource will be accessible as
        :return: the resource handle
        :raises asphalt.core.context.ResourceConflict: if there is an existing resource creator for
            the given types or context variable

        """
        assert check_argument_types()
        assert callable(creator), 'creator must be callable'
        return self._publish_resource(None, alias, context_attr, types, creator)

    def remove_resource(self, resource: Resource):
        """
        Remove the given resource from the collection and dispatch a ``resource_removed`` event.

        :param resource: the resource to be removed
        :raises LookupError: the given resource was not in the collection

        """
        assert check_argument_types()
        try:
            for typename in resource.types:
                del self._resources[typename][resource.alias]
        except KeyError:
            raise LookupError('{!r} not found in this context'.format(resource)) from None

        # Remove the creator from the resource creators
        if resource.creator is not None:
            del self._lazy_resources[resource.context_attr]

        # Remove the attribute from this context
        if resource.context_attr and resource.context_attr in self.__dict__:
            delattr(self, resource.context_attr)

        self.resource_removed.dispatch(resource)

    async def request_resource(self, type: Union[str, type], alias: str = 'default', *,
                               timeout: Union[int, None] = None):
        """
        Request a resource matching the given type and alias.

        If no such resource was found, this method will wait ``timeout`` seconds for it to become
        available. The timeout does not apply to resolving awaitables created by lazy resource
        creators.

        :param type: type of the requested resource
        :param alias: alias of the requested resource
        :param timeout: the timeout (in seconds; omit to use the context's default timeout)
        :return: the value contained by the requested resource (**NOT** a :class:`Resource`
            instance)
        :raises asphalt.core.context.ResourceNotFound: if the requested resource does not become
            available in the allotted time

        """
        assert check_argument_types()
        if not type:
            raise ValueError('type must be a type or a nonempty string')
        if not alias:
            raise ValueError('alias must be a nonempty string')

        resource_type = qualified_name(type) if not isinstance(type, str) else type
        timeout = timeout if timeout is not None else self.default_timeout
        assert timeout >= 0, 'timeout must be a positive integer'

        # Build a context chain from this context and its parents
        context_chain = [self]
        while context_chain[-1]._parent:
            context_chain.append(context_chain[-1]._parent)

        # First try to look up the resource in the context chain
        for ctx in context_chain:
            resource = ctx._resources.get(resource_type, {}).get(alias)
            if resource is not None:
                value = resource.get_value(self)
                return await value if isawaitable(value) else value

        # Listen to resource publish events in the whole chain and wait for the right kind of
        # resource to be published
        def resource_listener(event: ResourceEvent):
            if event.resource.alias == alias and resource_type in event.resource.types:
                future.set_result(event.resource)

        future = Future()
        for ctx in context_chain:
            ctx.resource_published.connect(resource_listener)

        try:
            resource = await wait_for(future, timeout)
        except TimeoutError:
            raise ResourceNotFound(resource_type, alias) from None
        else:
            value = resource.get_value(self)
            return await value if isawaitable(value) else value
        finally:
            for ctx in context_chain:
                ctx.resource_published.disconnect(resource_listener)

    def get_resources(self, type: Union[str, type] = None, *,
                      include_parents: bool = True) -> Sequence[Resource]:
        """
        Return the currently published resources specific to one type or all types.

        :param type: type of the resources to return, or ``None`` to return all resources
        :param include_parents: include the resources from parent contexts

        """
        resources = set(chain(*(value.values() for value in self._resources.values())))
        if include_parents and self._parent:
            resources = resources.union(self._parent.get_resources(type, include_parents=True))

        if type is not None:
            resource_type = qualified_name(type) if not isinstance(type, str) else type
            resources = (resource for resource in resources if resource_type in resource.types)

        return sorted(resources, key=lambda resource: (resource.types, resource.alias))
